# =========================================
# elixir dining philos
# =========================================

defmodule DPhil do
  def start do
    forks = Enum.map(1..3, fn _ -> spawn(&fork/0) end)

    Enum.each(1..3, fn i ->
      fork_left = List.at(forks, i - 1)
      fork_right = List.at(forks, rem(i, 3))

      spawn(fn -> philosopher(i, fork_left, fork_right) end)
    end)
  end

  defp fork() do
    receive do
      {:pick_up, pid} ->
        send(pid, :picked_up)
        fork()
      {:put_down, _pid} ->
        fork()
    end
  end

  defp philosopher(id, fork_left, fork_right) do
    loop(id, fork_left, fork_right)
  end

  defp loop(id, fork_left, fork_right) do
    think(id)

    fork_left |> send({:pick_up, self()})
    receive do
      :picked_up ->
        fork_right |> send({:pick_up, self()})
        receive do
          :picked_up ->
            eat(id)
            fork_right |> send({:put_down, self()})
        end
        fork_left |> send({:put_down, self()})
    end

    loop(id, fork_left, fork_right)
  end

  defp think(id) do
    IO.puts("Philosopher #{id} is thinking.")
    :timer.sleep(:rand.uniform(1000))
  end

  defp eat(id) do
    IO.puts("Philosopher #{id} is eating.")
    :timer.sleep(:rand.uniform(1000))
  end
end

# DPhil.start()

