
defmodule Philosopher do
  def start(id, left_fork, right_fork) do
    spawn(fn -> loop(id, left_fork, right_fork) end)
  end

  defp loop(id, left_fork, right_fork) do
    think(id)
    pick_forks(id, left_fork, right_fork)
    eat(id)
    put_forks(id, left_fork, right_fork)

    loop(id, left_fork, right_fork)
  end

  defp think(id) do
    IO.puts("Philosopher #{id} is thinking.")
    :timer.sleep(:rand.uniform(1000))
  end

  defp pick_forks(id, left_fork, right_fork) do
    send(left_fork, {:pick, id})
    send(right_fork, {:pick, id})
  end

  defp eat(id) do
    IO.puts("Philosopher #{id} is eating.")
    :timer.sleep(:rand.uniform(1000))
  end

  defp put_forks(id, left_fork, right_fork) do
    send(left_fork, {:put, id})
    send(right_fork, {:put, id})
  end
end

defmodule Fork do
  def start do
    spawn(fn -> loop() end)
  end

  defp loop() do
    receive do
      {:pick, philosopher_id} ->
        IO.puts("Philosopher #{philosopher_id} picked up a fork.")
        loop()

      {:put, philosopher_id} ->
        IO.puts("Philosopher #{philosopher_id} put down a fork.")
        loop()
    end
  end
end

defmodule DPhils do
  def start(num_philosophers) do
    forks = Enum.map(1..num_philosophers, fn _ -> Fork.start() end)

    Enum.each(1..num_philosophers, fn id ->
      left_fork = Enum.at(forks, rem(id - 1, num_philosophers))
      right_fork = Enum.at(forks, rem(id, num_philosophers))
      Philosopher.start(id, left_fork, right_fork)
    end)
  end
end

# Start the simulation with 5 philosophers
DPhils.start(5)

